import ast
import inspect
import re
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from types import GenericAlias, LambdaType, ModuleType, NoneType
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    Any,
    List,
    get_args,
    get_origin,
)

from cadwyn._compat import (
    PYDANTIC_V2,
    is_pydantic_1_constrained_type,
)
from cadwyn._package_utils import (
    get_absolute_python_path_of_import,
)
from cadwyn._utils import PlainRepr, UnionType
from cadwyn.exceptions import CodeGenerationError, InvalidGenerationInstructionError, ModuleIsNotAvailableAsTextError

if TYPE_CHECKING:
    import annotated_types

_LambdaFunctionName = (lambda: None).__name__  # pragma: no branch
_RE_CAMEL_TO_SNAKE = re.compile(r"(?<!^)(?=[A-Z])")


# A parent type of typing._GenericAlias
_BaseGenericAlias = type(List[int]).mro()[1]  # noqa: UP006

# type(list[int]) and type(List[int]) are different which is why we have to do this.
# Please note that this problem is much wider than just lists which is why we use typing._BaseGenericAlias
# instead of typing._GenericAlias.
GenericAliasUnion = GenericAlias | _BaseGenericAlias

_LambdaFunctionName = (lambda: None).__name__  # pragma: no branch


def get_fancy_repr(value: Any):
    if PYDANTIC_V2:
        import annotated_types

        if isinstance(value, annotated_types.GroupedMetadata) and hasattr(type(value), "__dataclass_fields__"):
            return transform_grouped_metadata(value)
    if isinstance(value, list | tuple | set | frozenset):
        return transform_collection(value)
    if isinstance(value, dict):
        return transform_dict(value)
    if isinstance(value, GenericAliasUnion):
        return transform_generic_alias(value)
    if value is None or value is NoneType:
        return transform_none(value)
    if isinstance(value, type):
        return transform_type(value)
    if isinstance(value, Enum):
        return transform_enum(value)
    if isinstance(value, auto):
        return transform_auto(value)
    if isinstance(value, UnionType):
        return transform_union(value)
    if isinstance(value, LambdaType) and _LambdaFunctionName == value.__name__:
        return transform_lambda(value)
    if inspect.isfunction(value):
        return transform_function(value)
    else:
        return transform_other(value)


def transform_grouped_metadata(value: "annotated_types.GroupedMetadata"):
    modified_fields = []
    empty_obj = type(value)

    for key in empty_obj.__dataclass_fields__:  # pyright: ignore[reportGeneralTypeIssues]
        if getattr(value, key) != getattr(empty_obj, key):
            modified_fields.append((key, getattr(value, key)))

    return PlainRepr(
        value.__class__.__name__
        + "("
        + ", ".join(f"{PlainRepr(key)}={get_fancy_repr(v)}" for key, v in modified_fields)
        + ")",
    )


def transform_collection(value: list | tuple | set | frozenset) -> Any:
    return PlainRepr(value.__class__(map(get_fancy_repr, value)))


def transform_dict(value: dict) -> Any:
    return PlainRepr(
        value.__class__((get_fancy_repr(k), get_fancy_repr(v)) for k, v in value.items()),
    )


def transform_generic_alias(value: GenericAliasUnion) -> Any:
    return f"{get_fancy_repr(get_origin(value))}[{', '.join(get_fancy_repr(a) for a in get_args(value))}]"


def transform_none(_: NoneType) -> Any:
    return "None"


def transform_type(value: type) -> Any:
    # This is a hack for pydantic's Constrained types
    if is_pydantic_1_constrained_type(value):
        parent = value.mro()[1]
        snake_case = _RE_CAMEL_TO_SNAKE.sub("_", value.__name__)
        cls_name = "con" + "".join(snake_case.split("_")[1:-1])
        return (
            cls_name.lower()
            + "("
            + ", ".join(
                [
                    f"{key}={get_fancy_repr(val)}"
                    for key, val in value.__dict__.items()
                    if not key.startswith("_") and val is not None and val != parent.__dict__[key]
                ],
            )
            + ")"
        )

    return value.__name__


def transform_enum(value: Enum) -> Any:
    return PlainRepr(f"{value.__class__.__name__}.{value.name}")


def transform_auto(_: auto) -> Any:
    return PlainRepr("auto()")


def transform_union(value: UnionType) -> Any:
    return "typing.Union[" + (", ".join(get_fancy_repr(a) for a in get_args(value))) + "]"


def transform_lambda(value: LambdaType) -> Any:
    # We clean source because getsource() can return only a part of the expression which
    # on its own is not a valid expression such as: "\n  .had(default_factory=lambda: 91)"
    return _get_lambda_source_from_default_factory(inspect.getsource(value).strip(" \n\t."))


def transform_function(value: Callable) -> Any:
    return PlainRepr(value.__name__)


def transform_other(value: Any) -> Any:
    return PlainRepr(repr(value))


def _get_lambda_source_from_default_factory(source: str) -> str:
    found_lambdas: list[ast.Lambda] = []

    ast.parse(source)
    for node in ast.walk(ast.parse(source)):
        if isinstance(node, ast.keyword) and node.arg == "default_factory" and isinstance(node.value, ast.Lambda):
            found_lambdas.append(node.value)
    if len(found_lambdas) == 1:
        return ast.unparse(found_lambdas[0])
    # These two errors are really hard to cover. Not sure if even possible, honestly :)
    elif len(found_lambdas) == 0:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            f"No lambda found in default_factory even though one was passed: {source}",
        )
    else:  # pragma: no cover
        raise InvalidGenerationInstructionError(
            "More than one lambda found in default_factory. This is not supported.",
        )


def read_python_module(module: ModuleType) -> str:
    # Can be cached in the future to gain some speedups
    try:
        return inspect.getsource(module)
    except OSError as e:
        if module.__file__ is None:  # pragma: no cover
            raise CodeGenerationError(f"Failed to get file path to the module {module}") from e
        path = Path(module.__file__)
        if path.is_file() and path.read_text() == "":
            return ""
        raise ModuleIsNotAvailableAsTextError(  # pragma: no cover
            f"Failed to get source code for module {module}. "
            "This is likely because this module is not available as code "
            "(it could be a compiled C extension or even a .pyc file). "
            "Cadwyn does not support models from such code. "
            "Please, open an issue on Cadwyn's issue tracker if you believe that your use case is valid "
            "and if you believe that it is possible for Cadwyn to support it.",
        ) from e


def get_all_names_defined_at_toplevel_of_module(body: ast.Module, module_python_path: str) -> dict[str, str]:
    """Some day we will want to use this to auto-add imports for new symbols in versions. Some day..."""
    defined_names = {}
    for node in body.body:
        if isinstance(node, ast.ClassDef | ast.FunctionDef | ast.AsyncFunctionDef):
            defined_names[node.name] = module_python_path
        elif isinstance(node, ast.Assign):
            for target in node.targets:
                if isinstance(target, ast.Name):
                    defined_names[target.id] = module_python_path
        elif isinstance(node, ast.ImportFrom):
            for name in node.names:
                defined_names[name.name] = get_absolute_python_path_of_import(node, module_python_path)
        elif isinstance(node, ast.Import):
            for name in node.names:
                defined_names[name.name] = name.name
    return defined_names


def add_keyword_to_call(attr_name: str, attr_value: Any, call: ast.Call):
    new_keyword = get_ast_keyword_from_argument_name_and_value(attr_name, attr_value)
    for i, keyword in enumerate(call.keywords):
        if keyword.arg == attr_name:
            call.keywords[i] = new_keyword
            break
    else:
        call.keywords.append(new_keyword)


def delete_keyword_from_call(attr_name: str, call: ast.Call):
    for i, keyword in enumerate(call.keywords):  # pragma: no branch
        if keyword.arg == attr_name:
            call.keywords.pop(i)
            break


def get_ast_keyword_from_argument_name_and_value(name: str, value: Any):
    if not isinstance(value, ast.AST):
        value = ast.parse(get_fancy_repr(value), mode="eval").body
    return ast.keyword(arg=name, value=value)


def pop_docstring_from_cls_body(cls_body: list[ast.stmt]) -> list[ast.stmt]:
    if (
        len(cls_body) > 0
        and isinstance(cls_body[0], ast.Expr)
        and isinstance(cls_body[0].value, ast.Constant)
        and isinstance(cls_body[0].value.value, str)
    ):
        return [cls_body.pop(0)]
    else:
        return []


@dataclass(slots=True)
class _ValidatorWrapper:
    func_ast: ast.FunctionDef
    index_of_validator_decorator: int
    field_names: set[str | ast.expr] | None
    is_deleted: bool = False


def get_validator_info_or_none(method: ast.FunctionDef) -> _ValidatorWrapper | None:
    for index, decorator in enumerate(method.decorator_list):
        # The cases we handle here:
        # * `Name(id="root_validator")`
        # * `Call(func=Name(id="validator"), args=[Constant(value="foo")])`
        # * `Attribute(value=Name(id="pydantic"), attr="root_validator")`
        # * `Call(func=Attribute(value=Name(id="pydantic"), attr="root_validator"), args=[])`

        if isinstance(decorator, ast.Call) and ast.unparse(decorator.func).endswith("validator"):
            if len(decorator.args) == 0:
                return _ValidatorWrapper(method, index, None)
            else:
                return _ValidatorWrapper(
                    method, index, {arg.value if isinstance(arg, ast.Constant) else arg for arg in decorator.args}
                )
        elif isinstance(decorator, ast.Name | ast.Attribute) and ast.unparse(decorator).endswith("validator"):
            return _ValidatorWrapper(method, index, None)
    return None



import ast
from collections.abc import Sequence
from typing import Annotated, Any, cast, get_args, get_origin

from typing_extensions import assert_never

from cadwyn._compat import (
    PYDANTIC_V2,
    FieldInfo,
    PydanticFieldWrapper,
    dict_of_empty_field_info,
    is_constrained_type,
)
from cadwyn._package_utils import IdentifierPythonPath, get_cls_pythonpath
from cadwyn._utils import Sentinel
from cadwyn.codegen._asts import add_keyword_to_call, delete_keyword_from_call, get_fancy_repr
from cadwyn.codegen._common import GlobalCodegenContext, PydanticModelWrapper, _EnumWrapper
from cadwyn.exceptions import InvalidGenerationInstructionError
from cadwyn.structure.enums import AlterEnumSubInstruction, EnumDidntHaveMembersInstruction, EnumHadMembersInstruction
from cadwyn.structure.schemas import (
    AlterSchemaSubInstruction,
    FieldDidntExistInstruction,
    FieldDidntHaveInstruction,
    FieldExistedAsInstruction,
    FieldHadInstruction,
    SchemaHadInstruction,
    ValidatorDidntExistInstruction,
    ValidatorExistedInstruction,
)


def class_migration_plugin(context: GlobalCodegenContext):
    for version_change in context.current_version.version_changes:
        _apply_alter_schema_instructions(
            context.schemas,
            version_change.alter_schema_instructions,
            version_change.__name__,
        )
        _apply_alter_enum_instructions(
            context.enums,
            version_change.alter_enum_instructions,
            version_change.__name__,
        )


def _apply_alter_schema_instructions(
    modified_schemas: dict[IdentifierPythonPath, PydanticModelWrapper],
    alter_schema_instructions: Sequence[AlterSchemaSubInstruction | SchemaHadInstruction],
    version_change_name: str,
):
    for alter_schema_instruction in alter_schema_instructions:
        schema = alter_schema_instruction.schema
        schema_path = get_cls_pythonpath(schema)
        schema_info = modified_schemas[schema_path]
        if isinstance(alter_schema_instruction, FieldExistedAsInstruction):
            _add_field_to_model(schema_info, modified_schemas, alter_schema_instruction, version_change_name)
        elif isinstance(alter_schema_instruction, FieldHadInstruction | FieldDidntHaveInstruction):
            _change_field_in_model(
                schema_info,
                modified_schemas,
                alter_schema_instruction,
                version_change_name,
            )
        elif isinstance(alter_schema_instruction, FieldDidntExistInstruction):
            _delete_field_from_model(schema_info, alter_schema_instruction.name, version_change_name)
        elif isinstance(alter_schema_instruction, ValidatorExistedInstruction):
            validator_name = alter_schema_instruction.validator.__name__
            schema_info.validators[validator_name] = alter_schema_instruction.validator_info
        elif isinstance(alter_schema_instruction, ValidatorDidntExistInstruction):
            if alter_schema_instruction.name not in schema_info.validators:
                raise InvalidGenerationInstructionError(
                    f'You tried to delete a validator "{alter_schema_instruction.name}" from "{schema_info.name}" '
                    f'in "{version_change_name}" but it doesn\'t have such a validator.',
                )
            if schema_info.validators[alter_schema_instruction.name].is_deleted:
                raise InvalidGenerationInstructionError(
                    f'You tried to delete a validator "{alter_schema_instruction.name}" from "{schema_info.name}" '
                    f'in "{version_change_name}" but it is already deleted.',
                )
            schema_info.validators[alter_schema_instruction.name].is_deleted = True
        elif isinstance(alter_schema_instruction, SchemaHadInstruction):
            _change_model(schema_info, alter_schema_instruction, version_change_name)
        else:
            assert_never(alter_schema_instruction)


def _apply_alter_enum_instructions(
    enums: dict[IdentifierPythonPath, _EnumWrapper],
    alter_enum_instructions: Sequence[AlterEnumSubInstruction],
    version_change_name: str,
):
    for alter_enum_instruction in alter_enum_instructions:
        enum = alter_enum_instruction.enum
        enum_path = get_cls_pythonpath(enum)
        enum = enums[enum_path]
        if isinstance(alter_enum_instruction, EnumDidntHaveMembersInstruction):
            for member in alter_enum_instruction.members:
                if member not in enum.members:
                    raise InvalidGenerationInstructionError(
                        f'You tried to delete a member "{member}" from "{enum.cls.__name__}" '
                        f'in "{version_change_name}" but it doesn\'t have such a member.',
                    )
                enum.members.pop(member)
        elif isinstance(alter_enum_instruction, EnumHadMembersInstruction):
            for member, member_value in alter_enum_instruction.members.items():
                if member in enum.members and enum.members[member] == member_value:
                    raise InvalidGenerationInstructionError(
                        f'You tried to add a member "{member}" to "{enum.cls.__name__}" '
                        f'in "{version_change_name}" but there is already a member with that name and value.',
                    )
                enum.members[member] = member_value
        else:
            assert_never(alter_enum_instruction)


def _change_model(
    model: PydanticModelWrapper,
    alter_schema_instruction: SchemaHadInstruction,
    version_change_name: str,
):
    # We only handle names right now so we just go ahead and check
    if alter_schema_instruction.name == model.name:
        raise InvalidGenerationInstructionError(
            f'You tried to change the name of "{model.name}" in "{version_change_name}" '
            "but it already has the name you tried to assign.",
        )

    model.name = alter_schema_instruction.name


def _add_field_to_model(
    model: PydanticModelWrapper,
    schemas: "dict[IdentifierPythonPath, PydanticModelWrapper]",
    alter_schema_instruction: FieldExistedAsInstruction,
    version_change_name: str,
):
    defined_fields = model._get_defined_fields_through_mro(schemas)
    if alter_schema_instruction.name in defined_fields:
        raise InvalidGenerationInstructionError(
            f'You tried to add a field "{alter_schema_instruction.name}" to "{model.name}" '
            f'in "{version_change_name}" but there is already a field with that name.',
        )

    fancy_type_repr = get_fancy_repr(alter_schema_instruction.type)
    field = PydanticFieldWrapper(
        annotation_ast=ast.parse(fancy_type_repr, mode="eval").body,
        annotation=alter_schema_instruction.type,
        init_model_field=alter_schema_instruction.field,
        value_ast=None,
    )
    model.fields[alter_schema_instruction.name] = field

    passed_field_attributes = field.passed_field_attributes
    if passed_field_attributes:
        field_call_ast = cast(ast.Call, ast.parse("Field()", mode="eval").body)
        for attr_name, attr_value in passed_field_attributes.items():
            add_keyword_to_call(attr_name, attr_value, field_call_ast)
        field.value_ast = field_call_ast
    model.cls.__annotations__[alter_schema_instruction.name] = alter_schema_instruction.type


def _change_field_in_model(
    model: PydanticModelWrapper,
    schemas: "dict[IdentifierPythonPath, PydanticModelWrapper]",
    alter_schema_instruction: FieldHadInstruction | FieldDidntHaveInstruction,
    version_change_name: str,
):
    defined_annotations = model._get_defined_annotations_through_mro(schemas)
    defined_fields = model._get_defined_fields_through_mro(schemas)
    if alter_schema_instruction.name not in defined_fields:
        raise InvalidGenerationInstructionError(
            f'You tried to change the field "{alter_schema_instruction.name}" from '
            f'"{model.name}" in "{version_change_name}" but it doesn\'t have such a field.',
        )

    field = defined_fields[alter_schema_instruction.name]
    model.fields[alter_schema_instruction.name] = field
    model.cls.__annotations__[alter_schema_instruction.name] = defined_annotations[alter_schema_instruction.name]

    annotation_ast, field_call_ast, contype_is_definitely_used = _get_constraint_asts_and_field_call_ast(
        schemas, model, alter_schema_instruction.name, field
    )

    constrained_type_annotation = None

    if annotation_ast is not None:
        field_type_is_annotated = True  # I.e. typing.Annotated
        # PydanticV2 changed field annotation handling so field.annotation lies to us
        real_annotation = model._get_defined_annotations_through_mro(schemas)[alter_schema_instruction.name]
        type_annotation = get_args(real_annotation)[0]
        if is_constrained_type(type_annotation):
            constrained_type_annotation = type_annotation
    else:
        field_type_is_annotated = False
        type_annotation = field.annotation
        if is_constrained_type(type_annotation):
            constrained_type_annotation = type_annotation
        annotation_ast = field.annotation_ast
    if field_call_ast is None:
        field_call_ast = field.value_ast

    if isinstance(alter_schema_instruction, FieldHadInstruction):
        # TODO: This naming sucks
        _change_field(
            model,
            alter_schema_instruction,
            version_change_name,
            defined_annotations,
            field,
            annotation_ast,
            field_call_ast,
            type_annotation,
            field_type_is_annotated,
            constrained_type_annotation,
        )
    else:
        _delete_field_attributes(
            model,
            alter_schema_instruction,
            version_change_name,
            field,
            annotation_ast,
            field_call_ast,
            type_annotation,
            constrained_type_annotation,
            contype_is_definitely_used,
        )


def _change_field(  # noqa: C901
    model: PydanticModelWrapper,
    alter_schema_instruction: FieldHadInstruction,
    version_change_name: str,
    defined_annotations: dict[str, Any],
    field: PydanticFieldWrapper,
    annotation_ast: ast.expr | None,
    field_call_ast: ast.expr | None,
    type_annotation: Any,
    field_type_is_annotated: bool,
    constrained_type_annotation: Any | None,
):
    if alter_schema_instruction.type is not Sentinel:
        if field.annotation == alter_schema_instruction.type:
            raise InvalidGenerationInstructionError(
                f'You tried to change the type of field "{alter_schema_instruction.name}" to '
                f'"{alter_schema_instruction.type}" from "{model.name}" in "{version_change_name}" '
                f'but it already has type "{field.annotation}"',
            )
        field.annotation = alter_schema_instruction.type
        model.cls.__annotations__[alter_schema_instruction.name] = alter_schema_instruction.type
        fancy_type_repr = get_fancy_repr(alter_schema_instruction.type)
        field.annotation_ast = ast.parse(fancy_type_repr, mode="eval").body

    if alter_schema_instruction.new_name is not Sentinel:
        if alter_schema_instruction.new_name == alter_schema_instruction.name:
            raise InvalidGenerationInstructionError(
                f'You tried to change the name of field "{alter_schema_instruction.name}" '
                f'from "{model.name}" in "{version_change_name}" '
                "but it already has that name.",
            )
        model.fields[alter_schema_instruction.new_name] = model.fields.pop(alter_schema_instruction.name)
        model.cls.__annotations__[alter_schema_instruction.new_name] = model.cls.__annotations__.pop(
            alter_schema_instruction.name,
            defined_annotations[alter_schema_instruction.name],
        )

    field_info = field.field_info

    dict_of_field_info = {k: getattr(field_info, k) for k in field_info.__slots__}
    if dict_of_field_info == dict_of_empty_field_info:
        field_info = FieldInfo()
        field.field_info = field_info
    for attr_name in alter_schema_instruction.field_changes.__dataclass_fields__:
        attr_value = getattr(alter_schema_instruction.field_changes, attr_name)
        if attr_value is not Sentinel:
            if field.passed_field_attributes.get(attr_name, Sentinel) == attr_value:
                raise InvalidGenerationInstructionError(
                    f'You tried to change the attribute "{attr_name}" of field '
                    f'"{alter_schema_instruction.name}" '
                    f'from "{model.name}" to {attr_value!r} in "{version_change_name}" '
                    "but it already has that value.",
                )
            if constrained_type_annotation is not None and hasattr(constrained_type_annotation, attr_name):
                _setattr_on_constrained_type(constrained_type_annotation, attr_name, attr_value)
                if isinstance(annotation_ast, ast.Call):
                    add_keyword_to_call(attr_name, attr_value, annotation_ast)
                elif not PYDANTIC_V2:  # pragma: no branch
                    field.update_attribute(name=attr_name, value=attr_value)
                    if isinstance(field_call_ast, ast.Call):  # pragma: no branch
                        add_keyword_to_call(attr_name, attr_value, field_call_ast)

            else:
                field.update_attribute(name=attr_name, value=attr_value)
                if field_type_is_annotated and attr_name == "default":
                    field.value_ast = ast.parse(get_fancy_repr(attr_value), mode="eval").body
                elif isinstance(field_call_ast, ast.Call):
                    add_keyword_to_call(attr_name, attr_value, field_call_ast)
                elif field.value_ast is not None:
                    field_call_ast = cast(ast.Call, ast.parse("Field()", mode="eval").body)
                    add_keyword_to_call("default", field.value_ast, field_call_ast)
                    add_keyword_to_call(attr_name, attr_value, field_call_ast)
                    field.value_ast = field_call_ast
                else:
                    field.value_ast = cast(ast.Call, ast.parse("Field()", mode="eval").body)
                    field_call_ast = field.value_ast
                    add_keyword_to_call(attr_name, attr_value, field_call_ast)


def _setattr_on_constrained_type(constrained_type_annotation: Any, attr_name: str, attr_value: Any) -> None:
    setattr(constrained_type_annotation, attr_name, attr_value)


def _delete_field_attributes(
    model: PydanticModelWrapper,
    alter_schema_instruction: FieldDidntHaveInstruction,
    version_change_name: str,
    field: PydanticFieldWrapper,
    type_annotation_ast: ast.expr | None,
    field_call_ast: ast.expr | None,
    type_annotation: Any,
    constrained_type_annotation: Any,
    contype_is_definitely_used: bool,
) -> None:
    for attr_name in alter_schema_instruction.attributes:
        if attr_name in field.passed_field_attributes:
            field.delete_attribute(name=attr_name)
            if isinstance(field_call_ast, ast.Call):
                delete_keyword_from_call(attr_name, field_call_ast)
            elif attr_name == "default":  # pragma: no branch
                field.value_ast = None
        # In case annotation_ast is a conint/constr/etc. Notice how we do not support
        # the same operation for **adding** constraints for simplicity.
        elif (hasattr(constrained_type_annotation, attr_name)) or contype_is_definitely_used:
            if hasattr(constrained_type_annotation, attr_name):
                _setattr_on_constrained_type(constrained_type_annotation, attr_name, None)
            if isinstance(type_annotation_ast, ast.Call):  # pragma: no branch
                delete_keyword_from_call(attr_name, type_annotation_ast)
        else:
            raise InvalidGenerationInstructionError(
                f'You tried to delete the attribute "{attr_name}" of field "{alter_schema_instruction.name}" '
                f'from "{model.name}" in "{version_change_name}" '
                "but it already doesn't have that attribute.",
            )


ContypeIsDefinitelyUsed = bool
CONTYPES = (
    "conbytes",
    "condate",
    "condecimal",
    "confloat",
    "conint",
    "conlist",
    "conset",
    "constr",
)


def _get_constraint_asts_and_field_call_ast(
    schemas: dict[IdentifierPythonPath, PydanticModelWrapper],
    model: PydanticModelWrapper,
    field_name: str,
    field: PydanticFieldWrapper,
) -> tuple[ast.expr | None, ast.Call | None, ContypeIsDefinitelyUsed]:
    """If the field type is Annotated and contains "Field" """
    # We return both annotation ast and field call ast because annotation might be a constrained type such as conint
    # and therefore contain constraints that we might want to remove.

    # ContypeIsDefinitely used is used to determine whether constr/conint/etc is used in Pydantic 2
    # because pydantic 2 changes original type hints to make sure that conint/constr/etc do not appear in annotations.

    real_annotation = model._get_defined_annotations_through_mro(schemas)[field_name]
    # typing.Annotated scenario
    if get_origin(real_annotation) == Annotated:
        index_of_field_info = _find_index_of_field_info_in_annotated(real_annotation)
        if not (isinstance(field.annotation_ast, ast.Subscript) and isinstance(field.annotation_ast.slice, ast.Tuple)):
            return (field.annotation_ast, None, True)

        unparsed_annotation = ast.unparse(field.annotation_ast.slice.elts[0])
        contype_is_definitely_used = any(contype in unparsed_annotation for contype in CONTYPES)

        # In pydantic 2, this means that in fact there is conint/constr/etc instead of an actual Annotated.
        # Yes, pydantic 2 changes original type hints to make sure that conint/constr/etc do not appear in types.

        if index_of_field_info is not None:
            field_call_ast = field.annotation_ast.slice.elts[index_of_field_info]

            return (field.annotation_ast.slice.elts[0], cast(ast.Call, field_call_ast), contype_is_definitely_used)
        return (field.annotation_ast.slice.elts[0], None, contype_is_definitely_used)
    return (None, None, False)


def _find_index_of_field_info_in_annotated(real_annotation: Any):
    # Pydantic turns `Annotated[conint(lt=2 + 5),                Field(default=11),     annotated_types.Gt(0)]` into:
    #                `Annotated[int, None, Interval(lt=7), None, FieldInfo(default=11), annotated_types.Gt(0)]`
    # Why? No idea. Probably due to its weird handling of constrained types. So we gotta go from the last element
    # because constrained types can only appear and mess up indexing in the first element.
    for i, arg in enumerate(reversed(get_args(real_annotation)), start=1):
        if isinstance(arg, FieldInfo):
            return -i
    return None


def _delete_field_from_model(model: PydanticModelWrapper, field_name: str, version_change_name: str):
    if field_name not in model.fields:
        raise InvalidGenerationInstructionError(
            f'You tried to delete a field "{field_name}" from "{model.name}" '
            f'in "{version_change_name}" but it doesn\'t have such a field.',
        )
    model.fields.pop(field_name)
    for validator_name, validator in model.validators.copy().items():
        if validator.field_names is not None and field_name in validator.field_names:
            validator.field_names.remove(field_name)

            validator_decorator = cast(
                ast.Call, validator.func_ast.decorator_list[validator.index_of_validator_decorator]
            )
            for arg in validator_decorator.args.copy():
                if isinstance(arg, ast.Constant) and arg.value == field_name:
                    validator_decorator.args.remove(arg)
            validator.func_ast.decorator_list[0]
            if not validator.field_names:
                model.validators[validator_name].is_deleted = True
